import sqlite3

import speech_recognition as sr
import google_speech
from .databases import Database


class Assistant:
    """
    Allows to build an assistant, which can:
    - Speak (in different languages)
    - Listen (in different languages)
    - Remember last recognised words
    - Recognise a word in last recognised words
    - Memorize keyword-value pairs
    - Remember values from keywords
    """

    def __init__(self, language='en', database_name='memory', memory_table='memory', listen_log_table='listen_log'):
        self.speak_language = language
        self.listen_language = language
        self.last_recognised = ''

        # Database connection and table creation
        self.DATABASE_NAME = database_name
        self.MEMORY_TABLE = memory_table
        self.LISTEN_LOG_TABLE = listen_log_table

        self.database = Database(database_name='./{}.sqlite'.format(self.DATABASE_NAME))
        self._create_initial_tables()

        # Recognizer and microphone initialization
        self.recognizer = sr.Recognizer()
        self.microphone = sr.Microphone()

        self.adjust_for_ambient_noise()
        # print('Energy: {}'.format(self.recognizer.energy_threshold))

    def _create_initial_tables(self):
        """Creates initial tables in the database."""

        tables = self.database.get_table_names()

        if self.MEMORY_TABLE not in tables:
            self.database.create_table(table_name=self.MEMORY_TABLE, columns={'key': 'TEXT', 'value': 'TEXT'})

        if self.LISTEN_LOG_TABLE not in tables:
            self.database.create_table(table_name=self.LISTEN_LOG_TABLE, columns={'sentence': 'TEXT'})

    def adjust_for_ambient_noise(self):
        """Listen to get a representative sample of the ambient noise and adjust to be more precise."""
        with self.microphone as source:
            self.recognizer.adjust_for_ambient_noise(source)

    def speak(self, text, language=None):
        speak_language = language if language else self.speak_language
        google_speech.main(text, speak_language, None)

    def listen(self, language=None):
        listen_language = language if language else self.listen_language
        recognised = ''

        with self.microphone as source:
            audio = self.recognizer.listen(source)
            try:
                recognised = self.recognizer.recognize_google(audio, language=listen_language)
                self.database.insert_into(table_name=self.LISTEN_LOG_TABLE, values={'sentence': recognised})
                print('"{}"'.format(recognised))
            except sr.UnknownValueError:
                # Google Speech Recognition could not understand audio
                print('Error: Assistant could not understand audio')
            except sr.RequestError as e:
                # Could not request results from Google STT
                print('Error: Could not request results from server; {}'.format(e))
            except Exception as e:
                print('Unknown error occurred: {}'.format(e))
            finally:
                self.last_recognised = recognised

    def has_heard(self, words):
        """
        The idea behind this function is to support a single word, a list of words or even a regex expression.
        Now regex is not supported.
        """
        if type(words) is str:
            return words in self.last_recognised
        else:
            return any(word in self.last_recognised for word in words)

    def memorize(self, keyword, value):
        """ Memorize a keyword value pair."""
        self.database.insert_into(table_name=self.MEMORY_TABLE, values={'key': keyword, 'value': value})

    def remember(self, keyword):
        """Returns the memorized value of the given keyword."""
        # TODO: This should use a 'Database' class method and not access directly to the cursor.
        try:
            self.database.cursor.execute('SELECT value FROM {} WHERE key="{}"'.format(self.MEMORY_TABLE, keyword))
        except sqlite3.OperationalError:  # Key doesn't exist
            return []

        result = self.database.cursor.fetchall()
        return [sentence for sublist in result for sentence in sublist]

    def remember_last(self, key):
        # TODO: This could be improved querying directly to database instead of calling 'remember' method.
        try:
            return self.remember(key)[-1]
        except IndexError:
            return ''

    def get_all_listened_words(self):
        # TODO: This should use a 'Database' class method and not access directly to the cursor.
        self.database.cursor.execute('SELECT * FROM {}'.format(self.LISTEN_LOG_TABLE))
        result = self.database.cursor.fetchall()
        return [sentence for sublist in result for sentence in sublist]

    def get_last_recognised_from_memory(self):
        # TODO: This could be improved querying directly to database instead of calling 'get_all_listened_words' method.
        try:
            return self.get_all_listened_words()[-1]
        except IndexError:
            return ''

    def forget_listened_words(self):
        # TODO: This should use a 'Database' class method and not access directly to the cursor.
        self.database.cursor.execute('DELETE FROM {}'.format(self.LISTEN_LOG_TABLE))



import sqlite3


class DatabaseException(Exception):
    pass


class Database:

    def __init__(self, database_name, backend='sqlite'):
        """ Now Database class only accepts 'sqlite' backend. However it's build in this way to allow more backends in
        the future."""
        self.database_name = database_name
        self.backend = backend
        self.connexion = sqlite3.connect(database_name)
        self.cursor = self.connexion .cursor()

    def _raise_not_implemented(self):
        raise NotImplementedError('Method not implemented for "{}" backend.'.format(self.backend))

    def get_database_name(self):
        """ Returns database_name"""
        return self.database_name

    def get_table_names(self):
        """ Return a list of the names of all the tables"""
        if self.backend == 'sqlite':
            self.cursor.execute("SELECT name FROM sqlite_master WHERE type='table'")
            result = self.cursor.fetchall()
            return [table for sublist in result for table in sublist]
        else:
            self._raise_not_implemented()

    def create_table(self, table_name, columns):
        """ Creates a table.

            table_name: string-name
            columns: dictionary with
        """
        try:
            if self.backend == 'sqlite':
                column_list = ['{} {}'.format(key, value) for key, value in columns.items()]
                initial = 'CREATE TABLE {} ({})'.format(table_name, ', '.join(column_list))
                self.cursor.execute(initial)
                self.connexion.commit()
            else:
                self._raise_not_implemented()
        except Exception as e:
            raise DatabaseException('There was an error trying to create the table "{}": "{}"'.format(table_name, e))

    def insert_into(self, table_name, values):
        """
        Insert values into a table.

            table_name: string-name
            values: dictionary. Example: {'key': keyword, 'value': value}
        """
        self.cursor.execute('INSERT INTO {} ({}) VALUES({})'.format(table_name, ', '.join(values.keys()),
                                                                    '"{}"'.format('", "'.join(values.values()))))
        self.connexion.commit()

    # TODO: Implement next methods:
    # def retrieve_from(self, table_name, )
    # delete_database()



from setuptools import setup

setup(name='voice_assistant',
      version='0.1.0',
      description='Python library to build a Voice Assistant',
      url='https://github.com/noviluni/voice_assistant',
      author='Marc H.',
      author_email='noviluni@gmail.com',
      license='MIT',
      packages=['voice_assistant'],
      install_requires=[
          'SpeechRecognition==3.8.1',
          'google-speech==1.0.16',
          'PyAudio==0.2.11',
      ],
      zip_safe=False)